Sprite = class('Sprite')

function Sprite:initialize(name)
	local o = {}
	setmetatable(o, self)
	self.name = name
	self.animation = {}
	return self
end

function Sprite:getTileSetName()
  return self.tileSet
end

--Pulls the start through start + length quads from the tile-batch array tile-set and stores the frames as an animation of sprite with the index name.
--Sprites should only be on one line
function Sprite:addAnimation(name, tileset, start, length)
	local dm = TileSize
	local i = 1
	self.animation[name] = {}
	self.tileSet = tileset
	while i <= length do
		local quad = start + i - 1
		self.animation[name][i] = Graphics.tilebatch[tileset][quad]
		i = i + 1
	end
end

function Sprite:getCurrentAnimation()
  return self.current
end

function Sprite:setFrame(anim, frame)
  assert(self.animation[anim][frame] ~= nil, "Attempted to set nonexistant frame")
  self.active = self.animation[anim][frame]
  self.current = anim
  self.length = #self.animation[anim]
  self.step = frame
end

function Sprite:startAnimation(name)
	assert(self.animation[name], "Animation " .. name .. " does not exist for " .. self.name .. ".")
	self.step = 1
	self.current = name
	self.length = #self.animation[name]
	self.active = self.animation[name][1]
end

function Sprite:getFrame()
  return self.active
end

function Sprite:getName()
  return self.name
end

function Sprite:nextFrame()
	assert(self.current, "No current animation for Sprite: " .. self.name .. ".")
	if self.step == self.length then
		self.step = 1
	else
		self.step = self.step + 1
	end
	self.active = self.animation[self.current][self.step]
end