--TODO: Write Unit Collision. Requires collision layer finished.

World = class('World')
--Format for Maps:
-- Level 6: unit (object)
-- Level 5: projectile (object)
-- Level 4: object (object)
-- Level 3: tileEffect (tile)
-- Level 2: collision (tile)
-- Level 1: tile (tile)

--Map Settings structure:
--size[x,y], name, mapPath, goals[...], scripts, tileSize 

function World:initialize(settings)
  self.size = settings.size
  self.tView = true
  self.view = 350
  self.name = settings.name
  self.mapPath = settings.mapPath
  self.goals = settings.goals
  self.map = STI.new(self.mapPath)
  self.map.cMap = self.map:getCollisionMap(2)
  self.scriptDir = settings.scripts
  self.tileSize = settings.tileSize
  Globals.World.Offset.x = (Globals.Window.w - self.size.x*TileSize)/2
  Globals.World.Offset.y = (Globals.Window.h - self.size.y*TileSize)/2
  self.lightWorld = nil
  self:initializeLightWorld()
end

function World:initializeLightWorld()
  self.lightWorld = love.light.newWorld()
  self.lightWorld.setAmbientColor(0,0,0)
  self.light = self.lightWorld:newLight(0, 0, 255, 255, 350)
  self.light.setGlowStrength(.01)
  self.light.setRange(self.view)
  self.light.setGlowSize(1)
  self.light.setColor(255, 255, 255) --]]
end

--Data Get/Set

function World:getUnit(id)
  return self.map.layers.unit[id]
end

function World:storeUnit(unit)
  local id = unit:getGUID()
  --If you're reading this you can see the issue first hand!
  --self.map.layers.unit[1]:getGUID() should return 1, thus why it is unit[1]. It however returns the last GUID added to any entity.
  if self.map.layers.unit[1] then
    assert(false, self.map.layers.unit[1]:getGUID())
  end
  self.map.layers.unit[id] = unit
  unit = {}
end

function World:getObject(id)
  return self.map.layers.object[id]
end

function World:storeObject(object)
  local id = object:getGUID()
  self.map.layers.object[id] = object
end

function World:getProjectile(id)
  return self.map.layers.projectile[id]
 end
 
 function World:storeProjectile(pjt)
  self.map.layers.projectile[pjt.GUID] = pjt
 end
 
function World:getUnitPosition(id)
  local x, y = self.map.layers.unit[id]:getPosition()
  return x, y
end

function World:getClosestUnit(x, y, units)
  local u = nil
  local d = nil
  for k, v in ipairs(units) do
    local tx, ty = v:getPosition()
    local dist = math.sqrt((x-tx)^2 + (y-ty)^2)
    if not u then
      u = k
      d = dist
    else
      if dist < d then
        u = k
        d = dist
      end
    end
  end
  return u
end

function World:getUnitsInRadius(x, y, r)
  local data = {}
  local i = 1
  for k, v in ipairs(self.map.layers.units) do
    local tx, ty = v:getPosition()
    if x - r >= tx and x + r <= tx and y - r >= ty and y + r <= ty then
      data[i] = k
      i = i + 1
    end
  end
end

function World:getUnitsInRegion(x1, x2, y1, y2)
  local data = {}
  local i = 1
  for k, v in ipairs(World.map.layers.units) do
    local tx, ty = v:getPosition()
    if tx >= x1 and tx <= x2 and ty >= y1 and ty <= y2 then
      data[i] = k
      i = i + 1
    end
  end
end

--Movement

function World:collision(x, y, dir, w, h)
  if dir == "left" then
    x = math.floor(x)
  elseif dir == "right" then
    x = math.ceil(x) + w
  elseif dir == "up" then
    y = math.floor(y)
  else
    y = math.ceil(y) + h
  end
  x = x + 1
  y = y + 1
  x = GFXN:round(x)
  y = GFXN:round(y)
  if self.map.cMap.data[y][x] == nil then
    assert(false, y .. " " .. x)
  end
  if self.map.cMap.data[y][x] == 1 then
    return true
  else
    return false
  end
end

--Refigure for half steps
function World:unitAlignGrid(id)
  local x, y = self:getUnitPosition(id)
  local decX = (x - math.floor(x))
  local decY = (y - math.floor(y))
  if (decX >= 0 and decX < .25) then
    x = math.floor(x)
  elseif (decX >= .25 and decX <= .5) or (decX <= .75 and decX > .5) then
    x = (math.floor(x) + .5)
  else
    x = math.ceil(x)
  end
  if (decY >= 0 and decY < .25) then
    y = math.floor(y)
  elseif (decY >= .25 and decY <= .5) or (decY <= .75 and decY > .5) then
    y = (math.floor(y) + .5)
  else
    y = math.ceil(y)
  end
  self:getUnit(id):moveTo(x, y)
end

function World:moveUnit(dir, id)
  local unit = self:getUnit(id)
  local dx, dy = 0, 0
  if dir == "left" then
    dx = -1
  elseif dir == "right" then
    dx = 1
  elseif dir == "up" then
    dy = -1
  else
    dy = 1
  end
  local x, y = unit:getPosition()
  unit.moving = true
  unit.move.tx = x + dx
  unit.move.ty = y + dy
  unit.move.dx = dx
  unit.move.dy = dy
  if unit.move.dir and unit.move.dir ~= dir then
    unit.move.changeDir = true
  end
  unit.move.dir = dir
  self:storeUnit(unit)
end

--Combat

function World:isInCombat(id)
  return self.map.layers.unit[id].combat
end

function World:newCombat()
  if self.map.newCombat then
    for k, v in ipairs(self.map.newCombat) do
      self.map.newCombat[k] = nil
      v:onEnterCombat()
    end
  end
end

function World:exitCombat(id)
  if self.map.layers.unit[id].combat == true then
    self.map.layers.unit[id].combat = nil
    self.map.oldCombat[id] = true
  end
end

function World:oldCombat()
  if self.map.oldCombat then
    for k, v in ipairs(self.map.oldCombat) do
      self.map.oldCombat = nil
      v:onExitCombat()
    end
  end
end

function World:enterCombat(id)
  if not self.map.layers.unit[id].combat then
    self.map.layers.unit[id].combat = true
    self.map.newCombat[id] = true
  else
	--New Syntax
    --Debug.updateMsg("Error", "Failed Combat")
  end
end

function World:checkAggro()
  for k, v in ipairs(self.map.layers.units) do
    if v.type == "mob" and v.faction == "hostile" and not v.combat then
      local x, y = v:getPosition()
      local uRad = self:getUnitsInRange(x, y, v.aggro)
      if #uRad > 1 then
        local t = self:getClosestUnit(x, y, uRad)
        self:registerAttack(k, t)
      elseif #uRad == 1 then
        self:registerAttack(k, uRad[1])
      end
      if uRad then
        self:enterCombat(k)
        if not self:isInCombat(t) then
          self:enterCombat(t)
        end
        self.aggro[k] = t
      end
    end
  end
end

function World:checkEvade()
  for k, v in ipairs(self.map.aggro) do
    local x, y = v:getPosition()
    if math.sqrt( (x - v.ix)^2 + (y - v.iy)^2 ) > v.aggro*2 then
      World.AI:reset(k)
    end
  end
end

--Main Functions, Updates

function World:updateUnits(dt)
  if #self.map.layers.unit >= 1 then
    for k, v in ipairs(self.map.layers.unit) do
      local x, y = v:getPosition()
      local speed = v:getSpeed()
      local dx, dy = v.move.dx, v.move.dy
      local dir = v.move.dir
      local id = v:getGUID()
      if dir then
        anim = dir .. "_move"
      end
      if v.moving then
        local mx = speed * dt * dx
        local my = speed * dt * dy
        mx = mx + x
        my = my + y
        if v.move.changeDir then
          self:unitAlignGrid(id)
          v.move.changeDir = false
          v.dSprite = 0
        end
        if self:collision(mx, my, dir, v.w, v.h) then
          v.moving = false
          self:unitAlignGrid(id)
          v.sprite:setFrame(anim, 1)
        else
          v:moveTo(mx, my)
          if v.sprite:getCurrentAnimation() ~= anim then
            v.sprite:startAnimation(anim)
          else
            v.dSprite = v.dSprite + dt
            if v.dSprite >= Globals.Draw.SpriteUpdate then
              v.dSprite = v.dSprite - Globals.Draw.SpriteUpdate
              v.sprite:nextFrame()
            end
          end
        end
      else
        self:unitAlignGrid(id)
        if not dir then
          v.sprite:setFrame("down_move", 1)
        else
          v.sprite:setFrame(anim, 1)
        end
      end
      --TODO: Add combat, add cast, add die.
      self:storeUnit(v)
    end
  end
end

function World:updateProjectiles(dt)
  if #self.map.layers.projectile >= 1 then
    for _, pjt in pairs(self.map.layers.projectile) do
      pjt.sprite:nextFrame()
      if pjt:move(dt) then
        pjt:finish()
        self.map.layers.projectile[pjt.GUID] = nil
      end
    end
  end
end

function World:draw()
  local oX, oY = Globals.World.Offset.x, Globals.World.Offset.y
  local ww = love.window.getWidth()
  local wh = love.window.getHeight()
  local x, y = self:getUnitPosition(Globals.PlayerUnitId)
  love.graphics.push()
  love.graphics.translate(oX, oY)
  self.map:setDrawRange(oX, oY, ww, wh)
  self.map:draw()
  self.light.setPosition(x*TileSize + .5*TileSize + oX, y*TileSize + .5*TileSize + oY)
  for _, v in ipairs(self.map.layers.unit) do
    if v.info.render then
      local x, y = v:getPosition()
      local dm = self.tileSize
      local name = v.sprite:getTileSetName()
      love.graphics.draw(Graphics.tileSheet[name], v.sprite:getFrame(), x*dm, y*dm)
      local yMod = -1
      local sN = v:getUnitSubName()
      love.graphics.setColor(0,0,0,255)
      if sN == nil then
        yMod = -.5
      else
        local sL = (string.len(sN) + 2)*12
        love.graphics.printf("<" .. sN .. ">", (x)*dm - .5*sL + dm, (y-.5)*dm, sL, "center")
      end
      local nL = (8 + string.len(tostring(v:getUnitLevel())) + string.len(v:getUnitName()))*12
      love.graphics.printf("(LvL. " .. v:getUnitLevel() .. ") " .. v:getUnitName(), (x)*dm - .5*nL + dm, (y-1)*dm, nL, "center")
    end
  end
  love.graphics.pop()
  self.lightWorld.update()
  self.lightWorld.drawShadow()
end

function World:resize(w, h)
  self.map:resize(w, h)
  self.lightWorld = nil
  self:initializeLightWorld()
  Globals.World.Offset.x = (Globals.Window.w - self.size.x*TileSize)/2
  Globals.World.Offset.y = (Globals.Window.h - self.size.y*TileSize)/2
end

function World:update(dt)
  self.map:update(dt)
  self:updateProjectiles(dt)
  self:updateUnits(dt)
end