function init()
  --All headers for sub-directories
  dH = {}
  dH.File = {}
  dH.Data = "data/"
  dH.Game = "src/game/"
  dH.Lib = "lib/"
  dH.Scripts = "src/scripts/"
  dH.GFX = dH.Data .. "graphics/"
  dH.Kbnd = dH.Data .. "Keybindings/"
  dH.Units = dH.Data .. "units/"
  dH.Maps = dH.Data .. "maps/"
  dH.MapData = dH.Data .. "mapData/"
  dH.UI = dH.Data .. "UI/"
  dH.File.Settings = "settings.lua"
  dH.Src = "src/"
  dH.SData = dH.Src .. "data/"

  --Load Globals and needed libraries
  class = require(dH.Lib .. "middleclass")
  require(dH.Data .. "globals")
  require(dH.Lib .. "tableFile")
  
  --Set Current tilesize to USIZE
  
  TileSize = Globals.World.UniversalTileSize

  --Initialize Settings

  if not love.filesystem.exists(dH.File.Settings) or RESETSETTINGS then
    settingFile = love.filesystem.newFile(dH.File.Settings)
    assert(settingFile:open("w"), "Couldn't open File: " .. dH.File.Settings)
    stg = require(dH.Data .. "settings")
    local tFile = tableFile:new()
    tFile:serialize(stg)
    settingFile:write(tFile.write)
    settingFile:close()
  else
    local fn = love.filesystem.load(dH.File.Settings)
    stg = fn()
  end
  Settings = stg

  --Load Screen based on settings
  local stg = Settings.system
  if stg.w == -1 or stg.h == -1 then
    stg.w, stg.h, _ = love.window.getMode()
  end
  love.window.setMode(stg.w, stg.h, stg.flags)
  love.resize(stg.w, stg.h)

  --All Internal and External System Modules
  require(dH.Lib .. "world")
  require(dH.Lib .. "Entity")
  STI = require(dH.Lib .. "STI")
  require(dH.Lib .. "Sprite")
  Debug = require(dH.Lib .. "hudebug")
  require(dH.Lib .. "GenFuncs")
  require(dH.Lib .. "Frames")
  require(dH.Lib .. "Keybinding")
  require(dH.Lib .. "light")
  require(dH.Lib .. "postshader")
  
  require(dH.SData .. "debug")
  require(dH.SData .. "draw")
  require(dH.SData .. "keys")
  require(dH.SData .. "UI")
  require(dH.SData .. "update")
  
--Initialize Pages for Debugging
  
  Debug.pageName(1, "Player Sprite")
  Debug.pageName(2, "Player Location")
  Debug.pageName(3, "Game Data")
  Debug.pageName(4, "Player Data")

--Game State Variable Initialization

  GameState = {}
  GameState.UI = true
  GameState.Game = true
  GameState.GameUI = false

  --Declare Media Holders
  
  Graphics = {}
  Audio = {}
  MapData = {}
  UI = {}
  Error = {}
  
  --Load Maps
  
  MapData.mapIndex = {}
  MapData[1] = require(dH.MapData .. "Tutorial")
  MapData.mapIndex["Tutorial"] = 1
  GameWorld = World:new(MapData[1])  
  
  --Load Keybindings
  
  local fxn = require(dH.Kbnd .. "internalFXN")
  local initBind = require(dH.Kbnd .. "initBindings")
  local keyNames = require(dH.Kbnd .. "keyNames")
  Keybindings.initialize(Settings.keybindings, fxn, keyNames)

  --Include graphics and fragment tile sheets

  Graphics.tileSheet = {}
  
  --TODO: Make tilesheets enumerable
  
  Graphics.tileSheet.unit = love.graphics.newImage(dH.GFX .. "test.png")
  Graphics.tilebatch = {}
  Graphics.tilebatch.unit = {}
  local w, h = Graphics.tileSheet.unit:getDimensions()
  local iter = 1
  local i, j = 0, 0
  local mTS = TileSize * 2
  while i <= h/mTS do
    while j <= w/mTS do
      Graphics.tilebatch.unit[iter] = love.graphics.newQuad(j*mTS, i*mTS, mTS, mTS, w, h)
      iter = iter + 1
      j = j + 1
    end
    j = 0
    i = i + 1
  end
  
  --Load Units (This is where the error is)
  
  --Load the GMTest Unit from file and pass it through Entity
  local tU = Entity:new(require(dH.Units .. "GMTest"))
  --Move the unit to the starting coordinates (found in data/maps/tutorial.lua
  tU:moveTo(GameWorld.map.properties.sX, GameWorld.map.properties.sY)
  tU:setGUID(1)
  tU:setUnitLevel(69)
  --Store the unit table into the GameWorld
  GameWorld:storeUnit(tU)
  tU = nil
  --Reload a second unit to tU
  tU = Entity:new(require(dH.Units .. "GMTest"))
  --Move to an arbitrary location
  tU:moveTo(2, 5)
  --Change the GUID to something arbitrary
  tU:setGUID(645969)
  tU:setUnitLevel(3003)
  --Store this unit
  GameWorld:storeUnit(tU)
  --UI Menu loading
  
	UI.escMenu = require(dH.UI .. "escMenu")
	UI.optMenu = require(dH.UI .. "optMenu")
  UI.gfxMenu = require(dH.UI .. "gfxMenu")
  
  --Load Key functions
  
  require(dH.SData .. "Keys")
  
end
